/*
 * (c) 2010-2019, SWD Embedded Systems Limited, http://www.kpda.ru
 */

// Portions Copyright (c) 1998 The NetBSD Foundation, Inc.
// All rights reserved.
//
// This code is derived from software contributed to The NetBSD Foundation
// by Lennart Augustsson (lennart@augustsson.net) at
// Carlstedt Research & Technology.
//
// Redistribution and use in source and binary forms, with or without
// modification, are permitted provided that the following conditions
// are met:
// 1. Redistributions of source code must retain the above copyright
//    notice, this list of conditions and the following disclaimer.
// 2. Redistributions in binary form must reproduce the above copyright
//    notice, this list of conditions and the following disclaimer in the
//    documentation and/or other materials provided with the distribution.
// 3. All advertising materials mentioning features or use of this software
//    must display the following acknowledgement:
//        This product includes software developed by the NetBSD
//        Foundation, Inc. and its contributors.
// 4. Neither the name of The NetBSD Foundation nor the names of its
//    contributors may be used to endorse or promote products derived
//    from this software without specific prior written permission.
//
// THIS SOFTWARE IS PROVIDED BY THE NETBSD FOUNDATION, INC. AND CONTRIBUTORS
// ``AS IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
// TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
// PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL THE FOUNDATION OR CONTRIBUTORS
// BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
// CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
// SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
// INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
// CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
// ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
// POSSIBILITY OF SUCH DAMAGE.

// Module Description:  keyboard class driver

#include <keyb.h>

keyb_ctrl_t				KeybCtrl;

int keyb_free( keyb_t *keyb )
{
	int		status;

	keyb->status	&= ~( KEYBOARD_PRESENT | KEYBOARD_IN_USE );
	if( keyb->ep_int ) {						// Stop the Interrupt In endpoint
		usbd_abort_pipe( keyb->ep_int );
	}

	pthread_sleepon_lock( );
	pthread_sleepon_signal( keyb->rbuf );
	pthread_sleepon_unlock( );

	if( status = rsrcdbmgr_devno_detach( keyb->attr.rdev, 0 ) ) {
		fprintf( stderr, "keyb_free:  rsrcddbmgr_devno_detach %d\n", errno );
	}
	if( status = resmgr_detach( KeybCtrl.dispatch, keyb->resmgr_id, _RESMGR_DETACH_ALL | _RESMGR_DETACH_CLOSE ) ) {
		fprintf( stderr, "keyb_free:  resmgr_detach %d\n", status );
	}
	usbd_free_urb( keyb->urb );
	usbd_free( keyb->ks_ndata );
	usbd_free( keyb->ks_odata );
	buff_destroy( keyb->rbuf );
	TAILQ_REMOVE( &KeybCtrl.dlist, keyb, dlink );
	if( status = usbd_detach( keyb->device ) ) {
		fprintf( stderr, "keyb_free:  usbd_detach %x\n", status );
	}
	return( status );
}

void keyb_signal_handler( )
{
	siginfo_t		info;
	sigset_t		set;
	keyb_t			*keyb;
	keyb_t			*nkeyb;

	sigfillset( &set );
	sigdelset( &set, SIGTERM );
	sigprocmask( SIG_BLOCK, &set, NULL );

	sigemptyset( &set );
	sigaddset( &set, SIGTERM );
	while( SignalWaitinfo( &set, &info ) == -1 )
		;

	for( keyb = TAILQ_FIRST( &KeybCtrl.dlist ); keyb; keyb = nkeyb ) {
		nkeyb = TAILQ_NEXT( keyb, dlink );
		keyb_free( keyb );
	}
}

int keyb_process_keys( keyb_t *keyb )
{
	keyb_report_t		*nd;
	keyb_report_t		*od;
	struct timespec		tv;
	_uint32				now;
	int					mod, omod;
	_uint16				ibuf[MAXKEYS];		// char events
	int					nkeys, i, j;
	int					key;
#define ADDKEY(c) ibuf[nkeys++] = (c)

	nd	= (keyb_report_t *)keyb->ks_ndata;
	od	= (keyb_report_t *)keyb->ks_odata;

	if( nd->keycode[0] == KEY_ERROR ) {
		return( EOK );			// ignore
	}

	clock_gettime( CLOCK_REALTIME, &tv );
	now = (_uint32)tv.tv_sec * 1000 + (_uint32)tv.tv_nsec / 1000000;

	nkeys	= 0;
	mod		= nd->modifiers;
	omod	= od->modifiers;
	if( mod != omod ) {
		for (i = 0; i < NMOD; i++) {
			if (( mod & ukbd_mods[i].mask) != (omod & ukbd_mods[i].mask ) ) {
				ADDKEY( ukbd_mods[i].key | ( mod & ukbd_mods[i].mask ? PRESS : RELEASE ) );
			}
		}
	}
	if( memcmp( nd->keycode, od->keycode, NKEYCODE) != 0 ) {
		// Check for released keys
		for( i = 0; i < NKEYCODE; i++) {
			key = od->keycode[i];
			if( key == 0 ) {
				continue;
			}
			for( j = 0; j < NKEYCODE; j++ ) {
				if( key == nd->keycode[j] ) {
					goto rfound;
				}
			}
			ADDKEY( key | RELEASE );
		rfound:
			;
		}
		
		/* Check for pressed keys. */
		for( i = 0; i < NKEYCODE; i++ ) {
			key = nd->keycode[i];
			if( key == 0 ) {
				continue;
			}
			for( j = 0; j < NKEYCODE; j++ ) {
				if( key == od->keycode[j] ) {
					goto pfound;
				}
			}
			ADDKEY( key | PRESS );
		pfound:
			;
		}
	}
	else {
		if( keyb->ks_nrep && now > keyb->ks_ntime ) {
//			buff_append( keyb->rbuf, keyb->ks_rep, 1 );
			buff_append( keyb->rbuf, keyb->ks_rep, keyb->ks_nrep );
			keyb->ks_ntime = now + 10;

			// wakeup anyone blocked
			pthread_sleepon_lock( );
			pthread_sleepon_signal( keyb->rbuf );
			pthread_sleepon_unlock( );
			return( EOK );
		}
	}

	*od = *nd;

	if( nkeys == 0 ) {
		return( EOK );
	}

	if( 1 ) {
		char cbuf[MAXKEYS * 2];
		int c;
		int npress;

		for( npress = i = j = 0; i < nkeys; i++ ) {
			key = ibuf[i];
			c = ukbd_trtab[key & CODEMASK];
			if( c == NN ) {
				continue;
			}

			// special case control break/pause
			if( c == 0x7f ) {
				if( ( mod & ( MOD_CONTROL_L | MOD_CONTROL_R ) ) ) {
					cbuf[j++] = 0xe0, cbuf[j++] = 0x46;
					cbuf[j++] = 0xe0, cbuf[j++] = 0xc6;
					break;
				}
				else {
					cbuf[j++] = 0xe1, cbuf[j++] = 0x1d, cbuf[j++] = 0x45;
					cbuf[j++] = 0xe1, cbuf[j++] = 0x9d,	cbuf[j++] = 0xc5;
					break;
				}
			}

			if( c & 0x80 ) {
				cbuf[j++] = 0xe0;
			}
			cbuf[j] = c & 0x7f;
			if( key & RELEASE ) {
				cbuf[j] |= 0x80;
			}
			else {
				// remember pressed keys for autorepeat
				if( c & 0x80 ) {
					keyb->ks_rep[npress++] = 0xe0;
				}
				keyb->ks_rep[npress++] = c & 0x7f;
			}
			j++;
		}
		buff_append( keyb->rbuf, cbuf, j );
		if( npress != 0 ) {
			keyb->ks_nrep	= npress;
			keyb->ks_ntime	= now + 200;
		}
		else {
			keyb->ks_nrep = 0;
		}

		// wakeup anyone blocked
		pthread_sleepon_lock( );
		pthread_sleepon_signal( keyb->rbuf );
		pthread_sleepon_unlock( );
	}
	return( EOK );
}

void keyb_int_cbf( struct usbd_urb *urb, struct usbd_pipe *pipe, void *hdl )
{
	keyb_t		*keyb;
	_uint32		ustatus;
	_uint32		alen;

	keyb = (keyb_t *)hdl;
	usbd_urb_status( urb, &ustatus, &alen );

	if( ( ustatus & USBD_URB_STATUS_MASK ) != USBD_STATUS_CMP ) {
		// check for a stall condition and clear it
		if( ( ustatus & USBD_USB_STATUS_MASK ) == USBD_STATUS_STALL ) {
			usbd_reset_pipe( pipe );
		}
		if( KeybCtrl.verbose ) {
			fprintf( stderr, "keyb_int_cbf:  ustatus %x\n", ustatus );
		}
	}
	else {
		keyb_process_keys( keyb );
	}

	if( keyb->status & KEYBOARD_IN_USE ) {		// re-hook interrupt endpoint
		usbd_setup_interrupt( keyb->urb, URB_DIR_IN, keyb->ks_ndata, keyb->ep_int_size );
		if( usbd_io( keyb->urb, keyb->ep_int, keyb_int_cbf, keyb, USBD_TIME_INFINITY ) ) {
			keyb->status &= ~( KEYBOARD_IN_USE | KEYBOARD_PRESENT );
		}
	}
}

int keyb_set_idle( keyb_t *keyb )
{
	_uint32			status;
	struct usbd_urb	*urb;
	_uint8			*buffer;
	int				retry=3;

	status = ENOMEM;

	if( !(urb = usbd_alloc_urb( NULL ) ) )
		return( status );
 	if ( (buffer = usbd_alloc( 1 ) ) ) {
		while ( retry-- ) {
			usbd_setup_vendor( urb, URB_DIR_OUT, 10, USB_TYPE_CLASS | USB_RECIPIENT_INTERFACE, 8 << 8, 0, NULL, 0 );
			if( ( status = usbd_io( urb, keyb->ep_cntl, NULL, keyb, USBD_TIME_INFINITY ) ) == EOK ) {
				// get idle and make sure it was set.. fix for IBM Sillitek keayboard
				usbd_setup_vendor( urb, URB_DIR_IN, 2, USB_TYPE_CLASS | USB_RECIPIENT_INTERFACE, 0, 0, buffer, 1 );
				if( ( ( status = usbd_io( urb, keyb->ep_cntl, NULL, keyb, USBD_TIME_INFINITY ) ) == EOK ) && (*buffer != 0) ) {
					break;
				}
			}
		}
		usbd_free( buffer );
	}

	usbd_free_urb( urb );

	return( status );
}

int keyb_set_report( keyb_t *keyb, _uint32 lchng )
{
	_uint32			status;
	struct usbd_urb	*urb;
	_uint8			*buffer;

	status = ENOMEM;
	if( urb = usbd_alloc_urb( NULL ) ) {
		if( buffer = usbd_alloc( 1 ) ) {
			*buffer = lchng;
			usbd_setup_vendor( urb, URB_DIR_OUT, HID_SET_REPORT, USB_TYPE_CLASS | USB_RECIPIENT_INTERFACE, HID_REPORT_OUT, keyb->instance.iface, buffer, 1 );
			if( ( status = usbd_io( urb, keyb->ep_cntl, NULL, keyb, USBD_TIME_INFINITY ) ) == EOK ) {
				keyb->leds = *buffer;
			}
			usbd_free_urb( urb );
		}
		usbd_free( buffer );
	}

	return( status );
}

int keyb_io_unblock( resmgr_context_t *ctp, io_pulse_t *msg, RESMGR_OCB_T *ocb )
{
	keyb_t		*keyb;

	keyb	= (keyb_t *)ocb->attr;

	pthread_sleepon_lock( );
	keyb->status |= KEYBOARD_UNBLOCK;
	pthread_sleepon_signal( keyb->rbuf );
	pthread_sleepon_unlock( );

	return( _RESMGR_NOREPLY );
}

int keyb_io_open( resmgr_context_t *ctp, io_open_t *msg, RESMGR_HANDLE_T *handle, void *extra )
{
	keyb_t		*keyb;
	int			status;

	keyb	= (keyb_t *)handle;

	if( msg->connect.ioflag & _IO_FLAG_RD ) {
		msg->connect.sflag = SH_DENYRD;
	}

	if( msg->connect.ioflag & _IO_FLAG_WR ) {
		msg->connect.sflag = SH_DENYWR;
	}

	if( status = iofunc_open_default( ctp, msg, handle, extra ) ) {
		return( status );
	}

	if(	!( keyb->status & KEYBOARD_IN_USE ) ) {		// hook callback
		usbd_setup_interrupt( keyb->urb, URB_DIR_IN, keyb->ks_ndata, keyb->ep_int_size );
		if( status = usbd_io( keyb->urb, keyb->ep_int, keyb_int_cbf, keyb, USBD_TIME_INFINITY ) ) {
			return( status );
		}
		keyb->status |= KEYBOARD_IN_USE;
	}

	return( EOK );
}

int keyb_io_close( resmgr_context_t *ctp, void *reserved, RESMGR_OCB_T *ocb )
{
	keyb_t		*keyb;

	keyb	= (keyb_t *)ocb->attr;

	iofunc_attr_lock( ocb->attr );
	if( iofunc_ocb_detach( ctp, ocb ) & IOFUNC_OCB_LAST_READER ) {
		keyb->status &= ~KEYBOARD_IN_USE;
	}
	iofunc_ocb_free( ocb );
	iofunc_attr_unlock( &keyb->attr );
	return( EOK );
}

int keyb_io_read( resmgr_context_t *ctp, io_read_t *msg, RESMGR_OCB_T *ocb )
{
	keyb_t		*keyb;
	buffer_t	*bptr;
	char		kbuf[1];
	int			pos;
	int			nbytes;

	keyb	= (keyb_t *)ocb->attr;
	bptr	= keyb->rbuf;
	pos		= 0;
	nbytes	= msg->i.nbytes;
	keyb->status &= ~KEYBOARD_UNBLOCK;

	if( !( keyb->status & KEYBOARD_PRESENT ) ) {
		return( ENODEV );
	}

	if( ( msg->i.xtype & _IO_XTYPE_MASK ) != _IO_XTYPE_NONE ) {
		return( EINVAL );
	}

	if( !( ocb->ioflag & _IO_FLAG_RD ) ) {
		return( EBADF );
	}

	ocb->flags |= IOFUNC_ATTR_ATIME;

	if( nbytes == 0 ) {
		_IO_SET_READ_NBYTES( ctp, 0 );
		return( EOK );
	}

	while( 1 ) {
// Fix:  optimization buff_delete more than one packet at a time
		while( nbytes && buff_delete( bptr, kbuf, 1 ) > 0 ) {
			resmgr_msgwrite( ctp, kbuf, 1, pos );
			nbytes--;
			pos++;
		}

		if( !nbytes || pos >= keyb->tios.c_cc[VMIN] ) {
			break;
		}
		pthread_sleepon_lock( );
		while( buff_waiting( bptr ) == 0 ) {
			if( keyb->status & KEYBOARD_UNBLOCK ) {
				pthread_sleepon_unlock( );
				return( EINTR );
			}
			if( !( keyb->status & KEYBOARD_PRESENT ) ) {
				pthread_sleepon_unlock( );
				return( ENODEV );
			}
			pthread_sleepon_wait( bptr );
		}
		pthread_sleepon_unlock( );
	}

	_IO_SET_READ_NBYTES( ctp, pos );
	return( EOK );
}

int keyb_io_devctl( resmgr_context_t *ctp, io_devctl_t *msg, RESMGR_OCB_T *ocb )
{
	keyb_t		*keyb;
	_uint32		*data;
	int			nbytes;
	int			status;

	nbytes	= 0;
	data	= (_uint32 *)_DEVCTL_DATA( msg->i );
	keyb	= (keyb_t *)ocb->attr;

	if( !( keyb->status & KEYBOARD_PRESENT ) ) {
		return( ENODEV );
	}

		// Let common code handle DCMD_ALL_* cases
	if( ( status = iofunc_devctl_default( ctp, msg, ocb ) ) != _RESMGR_DEFAULT ) {
		return( status );
	}

	status = EOK;

	switch( msg->i.dcmd ) {
		case DCMD_CHR_TCGETATTR:
			nbytes = sizeof( struct termios );
			memcpy( data, &keyb->tios, nbytes );
			break;

		case DCMD_CHR_TCSETATTR:
			nbytes = sizeof( struct termios );
			memcpy( &keyb->tios, data, nbytes );
			break;

		case DCMD_CHR_TCFLUSH:
			buff_flush( keyb->rbuf );
			break;

		case DCMD_CHR_TTYINFO: {
			struct _ttyinfo			*ttyinfo;

			ttyinfo = (struct _ttyinfo *)data;
			strcpy( ttyinfo->ttyname, "kbd" );
			ttyinfo->opencount = keyb->attr.count;
			nbytes = sizeof( *ttyinfo );
			break;
		}

		case DCMD_CHR_LINESTATUS:
			if( keyb->leds & SCROLLLOCK_MAP ) {
				*data |= _LINESTATUS_CON_SCROLL;
			}
			if( keyb->leds & NUMLOCK_MAP ) {
				*data |= _LINESTATUS_CON_NUM;
			}
			if( keyb->leds & CAPSLOCK_MAP ) {
				*data |= _LINESTATUS_CON_CAPS;
			}
			nbytes = sizeof( *data );
			break;

		case DCMD_CHR_SERCTL: {
			_uint32	leds = 0;

			if( *data & _CONCTL_NUM )		leds |= NUMLOCK_MAP;
			if( *data & _CONCTL_CAPS )		leds |= CAPSLOCK_MAP;
			if( *data & _CONCTL_SCROLL )	leds |= SCROLLLOCK_MAP;
			keyb_set_report( keyb, leds );
			break;
		}

		default:
			status = ENOSYS;
			break;
	}

	msg->o.ret_val = status;
	return( _RESMGR_PTR( ctp, &msg->o, sizeof( msg->o ) + nbytes ) );
}

int keyb_io_lock_ocb( resmgr_context_t *ctp, void *reserved, RESMGR_OCB_T *ocb )
{
	return( EOK );
}

int keyb_io_unlock_ocb( resmgr_context_t *ctp, void *reserved, RESMGR_OCB_T *ocb )
{
	return( EOK );
}

int keyb_get_devnum( )
{
	int				cnt;
	int				match;
	keyb_t			*keyb;

	for( cnt = 0; cnt < 16; cnt++ ) {
		for( match = 0, keyb = TAILQ_FIRST( &KeybCtrl.dlist ); keyb; keyb = TAILQ_NEXT( keyb, dlink ) ) {
			if( minor( keyb->devnum ) == cnt ) {
				match = 1;
				break;
			}
		}
		if( !match ) {
			break;
		}
	}
if( cnt ) {
	fprintf( stderr, "get_devnum\n" );
}
	return( cnt );
}

int keyb_resmgr_attach( keyb_t *keyb )
{
	char		name[255];

	// setup termios structure
	keyb->tios.c_cc[VMIN] = 1;
	iofunc_attr_init( &keyb->attr, 0600 | S_IFCHR, 0, 0 );
	iofunc_func_init( _RESMGR_CONNECT_NFUNCS, &keyb->resmgr_connect_funcs, _RESMGR_IO_NFUNCS, &keyb->resmgr_io_funcs );
	if( ( keyb->attr.rdev = rsrcdbmgr_devno_attach( "usb", -1, 0 ) ) != -1 ) {
		keyb->devnum = keyb_get_devnum( );
		sprintf( name, "/dev/%s%d", KeybCtrl.prefix, keyb->devnum );
		keyb->resmgr_connect_funcs.open		= keyb_io_open;
		keyb->resmgr_io_funcs.close_ocb		= keyb_io_close;
		keyb->resmgr_io_funcs.unlock_ocb	= keyb_io_unlock_ocb;
		keyb->resmgr_io_funcs.lock_ocb		= keyb_io_lock_ocb;
		keyb->resmgr_io_funcs.read			= keyb_io_read;
		keyb->resmgr_io_funcs.devctl		= keyb_io_devctl;
		keyb->resmgr_io_funcs.unblock		= keyb_io_unblock;
		if( ( keyb->resmgr_id = resmgr_attach( KeybCtrl.dispatch, NULL, name, _FTYPE_ANY, 0,
				&keyb->resmgr_connect_funcs, &keyb->resmgr_io_funcs, &keyb->attr ) ) != -1 ) {
			resmgr_devino( keyb->resmgr_id, &keyb->io_mount.dev, &keyb->attr.inode );
			if( ( keyb->rbuf = buff_create( 0, 64, 1 ) ) ) {
				return( EOK );
			}
			else {
				fprintf( stderr, "Keyb:  cannot alloc ring buffer\n" );
			}
		}
		else {
			fprintf( stderr, "Keyb:  resmgr_attach\n" );
		}
		rsrcdbmgr_devno_detach( keyb->attr.rdev, 0 );
	}
	else {
		fprintf( stderr, "Unable to acquire devno\n" );
	}

	return( EIO );
}

int keyb_parse_descriptors( keyb_t *keyb )
{
	_uint32						eix;
	_uint32						scan;
	_uint32						found;
	usbd_descriptors_t			*desc;
	struct usbd_desc_node		*ifc, *ept;

	scan	= KEYBOARD_CONTROL_EP | KEYBOARD_INTIN_EP;
	found	= 0;

	if( usbd_interface_descriptor( keyb->device, keyb->instance.config, keyb->instance.iface, keyb->instance.alternate, &ifc ) ) {
		for( eix = 0; ( desc = usbd_parse_descriptors( keyb->device, ifc, USB_DESC_ENDPOINT, eix, &ept ) ) != NULL; ++eix ) {
			switch( desc->endpoint.bmAttributes ) {
				case USB_ATTRIB_CONTROL:
					if( usbd_open_pipe( keyb->device, desc, &keyb->ep_cntl ) == EOK ) {
						found |= KEYBOARD_CONTROL_EP;
					}
					break;

				case USB_ATTRIB_ISOCHRONOUS:
					break;

				case USB_ATTRIB_BULK:
					break;

				case USB_ATTRIB_INTERRUPT:
					switch( desc->endpoint.bEndpointAddress & USB_ENDPOINT_IN ) {
						case USB_ENDPOINT_OUT:
							break;

						case USB_ENDPOINT_IN:
							if( usbd_open_pipe( keyb->device, desc,
									&keyb->ep_int ) == EOK ) {
								keyb->ep_int_size = desc->endpoint.wMaxPacketSize;
								found |= KEYBOARD_INTIN_EP;
							}
							break;
					}
					break;
			}
		}
	}
	return( ( found == scan ) ? EOK : ENODEV );
}

void keyb_insertion( struct usbd_connection *connection, usbd_device_instance_t *instance )
{
	keyb_t						*keyb;
	struct usbd_device			*device;
	int							status;

	if( ( status = usbd_attach( connection, instance, sizeof( keyb_t ), &device ) ) == EOK ) {
		keyb				= usbd_device_extra( device );
		keyb->device		= device;
		keyb->instance		= *instance;
		if( keyb_parse_descriptors( keyb ) == EOK ) {
			if( keyb->urb = usbd_alloc_urb( NULL ) ) {
				if( ( keyb->ks_ndata = usbd_alloc( keyb->ep_int_size ) ) ) {
					if( ( keyb->ks_odata = usbd_alloc( keyb->ep_int_size ) ) ) {
						keyb_set_idle( keyb );
						if( keyb_resmgr_attach( keyb ) == EOK ) {
							keyb->status = KEYBOARD_PRESENT;
							TAILQ_INSERT_TAIL( &KeybCtrl.dlist, keyb, dlink );
							return;
						}
						usbd_free( keyb->ks_odata );
					}
					usbd_free( keyb->ks_ndata );
				}
				usbd_free_urb( keyb->urb );
			}
		}
		usbd_detach( device );
	}
}

void keyb_removal( struct usbd_connection *connection, usbd_device_instance_t *instance )
{
	struct usbd_device	*device;
	keyb_t				*keyb;

	if( ( device = usbd_device_lookup( connection, instance ) ) != NULL ) {
		keyb = usbd_device_extra( device );
		keyb_free( keyb );
	}
}

int keyb_resmgr_init( )
{
	resmgr_attr_t		attr;

	if( procmgr_daemon( EXIT_SUCCESS, PROCMGR_DAEMON_NOCLOSE | PROCMGR_DAEMON_NODEVNULL ) == -1 ) {
		fprintf( stderr, "Keyb:  procmgr_daemon\n" );
		return( -1 );
	}

	if( ( KeybCtrl.dispatch = dispatch_create( ) ) == NULL ) {
		fprintf( stderr, "Unable to allocate dispatch context\n" );
		return( -1 );
	}

	memset( &KeybCtrl.pool_attr, 0, sizeof( KeybCtrl.pool_attr ) );
	if( ( KeybCtrl.pool_attr.attr = malloc( sizeof( *KeybCtrl.pool_attr.attr ) ) ) == NULL ) {
		fprintf( stderr, "Unable to alloc pool attr\n" );
		return( -1 );
	}

	pthread_attr_init( KeybCtrl.pool_attr.attr );
	pthread_attr_setstacksize( KeybCtrl.pool_attr.attr, 8192 );

	KeybCtrl.pool_attr.handle			= KeybCtrl.dispatch;
	KeybCtrl.pool_attr.context_alloc	= resmgr_context_alloc;
	KeybCtrl.pool_attr.block_func		= resmgr_block;
	KeybCtrl.pool_attr.handler_func		= (void *)dispatch_handler;
	KeybCtrl.pool_attr.context_free		= resmgr_context_free;
	KeybCtrl.pool_attr.lo_water			= 1;
	KeybCtrl.pool_attr.hi_water			= 2;
	KeybCtrl.pool_attr.increment		= 1;
	KeybCtrl.pool_attr.maximum			= 2;

	if( ( KeybCtrl.thread_pool = thread_pool_create( &KeybCtrl.pool_attr, 0 ) ) == NULL ) {
		fprintf( stderr, "Unable to initialize thread pool\n" );
		return( -1 );
	}

	memset( &attr, 0, sizeof( attr ) );
	attr.nparts_max		= 10;
	attr.msg_max_size	= 0;

	if( resmgr_attach( KeybCtrl.dispatch, &attr, NULL, 0, 0, NULL, NULL, NULL ) == -1 ) {
		fprintf( stderr, "Unable to attach name\n" );
		return( -1 );
	}

	if( thread_pool_start( KeybCtrl.thread_pool ) ) {
		fprintf( stderr, "Unable to initialize thread_pool\n" );
		return( -1 );
	}

	return( 0 );
}

int keyb_start_driver( int argc, char *argv[] )
{
	struct usbd_connection	*connection;
	usbd_funcs_t			funcs = { _USBDI_NFUNCS, keyb_insertion, keyb_removal, NULL };
	usbd_device_ident_t		interest = { USBD_CONNECT_WILDCARD, USBD_CONNECT_WILDCARD, USB_CLASS_HID, USB_SUBCLASS_HID_BOOT, USB_PROTOCOL_KEYBOARD };
	usbd_connect_parm_t		parm = { NULL, USB_VERSION, USBD_VERSION, 0, argc, argv, 0, &interest, &funcs, USBD_CONNECT_WAIT };
	int						status;
	int						opt;
	sigset_t				signals;

	sigfillset( &signals );					// No signals to resmgr threads
	pthread_sigmask( SIG_BLOCK, &signals, NULL );

	TAILQ_INIT( &KeybCtrl.dlist );
	KeybCtrl.prefix = "usbkbd";

	while( ( opt = getopt( argc, argv, "vn:s:w:" ) ) != -1 ) {
		switch( opt ) {
			case 'n':
				KeybCtrl.prefix = optarg;
				break;

			case 'v':
				KeybCtrl.verbose++;
				break;

			case 's':
				parm.path = optarg;
				break;

			case 'w':
				parm.connect_wait = strtol( optarg, 0, 0 );
				break;

			default:
				break;
		}
	}

	if( ( KeybCtrl.chid = ChannelCreate( _NTO_CHF_DISCONNECT | _NTO_CHF_UNBLOCK ) ) == -1 ||
	   ( KeybCtrl.coid = ConnectAttach( 0, 0, KeybCtrl.chid, _NTO_SIDE_CHANNEL, 0 ) ) == -1 ) {
		fprintf( stderr, "Unable to attach channel and connection\n" );
		return( errno );
	}

	if( keyb_resmgr_init( ) ) {
		fprintf( stderr, "Keyb:  Resmgr Init Failure\n" );
		return( errno );
	}

	if( ( status = usbd_connect( &parm, &connection ) ) != EOK ) {
		fprintf( stderr, "Keyb Class/init - %s\n", strerror( status ) );
		return( status );
	}

	return( status );
}

int main( int argc, char *argv[] )
{
	if( keyb_start_driver( argc, argv ) ) {
		exit( EXIT_FAILURE );
	}
	keyb_signal_handler( );
	exit( EXIT_SUCCESS );
}
