/*
 * (c) 2010-2019, SWD Embedded Systems Limited, http://www.kpda.ru
 */

// Module Description:  header for keyboard class driver

#include <share.h>
#include <errno.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <termios.h>
#include <gulliver.h>
#include <sys/usbdi.h>
#include <sys/iofunc.h>
#include <sys/resmgr.h>
#include <sys/procmgr.h>
#include <sys/dcmd_chr.h>
#include <sys/dispatch.h>
#include <sys/neutrino.h>
#include <sys/rsrcdbmgr.h>

// private headers
#include <queue.h>

typedef struct buffer buffer_t;

#define BUFF_OVERWRITE		0x01
struct buffer {
	int			head;
	int			tail;
	int			bufsize;
	int			recsize;
	char		*ptr;
	unsigned	mode;
	int			min;
};

//                    KEYBOARD BUFFER REPORT DEFINITION
#define NKEYCODE 6
#define NMOD 8

#define PRESS		0x000
#define RELEASE		0x100
#define CODEMASK	0x0ff

#define KEY_ERROR	0x01
#define MAXKEYS		(NMOD+2*NKEYCODE)

#define MOD_CONTROL_L	0x01
#define MOD_CONTROL_R	0x10
#define MOD_SHIFT_L		0x02
#define MOD_SHIFT_R		0x20
#define MOD_ALT_L		0x04
#define MOD_ALT_R		0x40
#define MOD_WIN_L		0x08
#define MOD_WIN_R		0x80

typedef struct _keyb_report {
	uint8_t	modifiers;
	uint8_t	reserved;
	uint8_t	keycode[NKEYCODE];
} keyb_report_t;

typedef struct _keyb {
	iofunc_attr_t					attr;
	TAILQ_ENTRY( _keyb )			dlink;
	int								resmgr_id;
	_uint32							devnum;
	resmgr_connect_funcs_t			resmgr_connect_funcs;
	resmgr_io_funcs_t				resmgr_io_funcs;
	iofunc_mount_t					io_mount;
	struct termios					tios;
	struct usbd_connection			*connection;
	usbd_device_instance_t			instance;
	struct usbd_device				*device;
	_uint32							status;
	struct usbd_pipe				*ep_cntl;
	struct usbd_pipe				*ep_int;
	uint32_t						ep_int_size;
	struct usbd_urb					*urb;
	_uint8							*ep_int_buf;
	buffer_t						*rbuf;			// ring buffer

	uint8_t							leds;
	uint8_t							rsvd[3];

	keyb_report_t					*ks_ndata;
	keyb_report_t					*ks_odata;
	uint32_t						ks_ntime;
	int								ks_nrep;
	char							ks_rep[MAXKEYS];
} keyb_t;

typedef struct _keyb_ctrl {
	TAILQ_HEAD(,_keyb)				dlist;		// linked list of mice
	_uint32							cflags;
	_uint32							verbose;
	char							*prefix;
	int								coid;
	int								chid;
	thread_pool_t 					*thread_pool;
	dispatch_t						*dispatch;
	thread_pool_attr_t				pool_attr;
} keyb_ctrl_t;

static struct {
	int mask, key;
} ukbd_mods[NMOD] = {
	{ MOD_CONTROL_L, 0xe0 },
	{ MOD_CONTROL_R, 0xe4 },
	{ MOD_SHIFT_L,   0xe1 },
	{ MOD_SHIFT_R,   0xe5 },
	{ MOD_ALT_L,     0xe2 },
	{ MOD_ALT_R,     0xe6 },
	{ MOD_WIN_L,     0xe3 },
	{ MOD_WIN_R,	 0xe7 },
};

#define NN 0		// no translation
// Translate USB keycodes to US keyboard AT/XT scancodes.
// Scancodes >= 128 represent EXTENDED keycodes.
static _uint8 ukbd_trtab[256] = {
	  NN,  NN,  NN,  NN,  30,  48,  46,  32, /* 00 - 07 */
	  18,  33,  34,  35,  23,  36,  37,  38, /* 08 - 0F */
	  50,  49,  24,  25,  16,  19,  31,  20, /* 10 - 17 */
	  22,  47,  17,  45,  21,  44,   2,   3, /* 18 - 1F */
	   4,   5,   6,   7,   8,   9,  10,  11, /* 20 - 27 */
	  28,   1,  14,  15,  57,  12,  13,  26, /* 28 - 2F */
	  27,  43,  43,  39,  40,  41,  51,  52, /* 30 - 37 */
	  53,  58,  59,  60,  61,  62,  63,  64, /* 38 - 3F */
	  65,  66,  67,  68,  87,  88, 183,  70, /* 40 - 47 */
	 127, 210, 199, 201, 211, 207, 209, 205, /* 48 - 4F */
	 203, 208, 200,  69, 181,  55,  74,  78, /* 50 - 57 */
	 156,  79,  80,  81,  75,  76,  77,  71, /* 58 - 5F */
      72,  73,  82,  83,  86, 221,  NN,  NN, /* 60 - 67 */
	  NN,  NN,  NN,  NN,  NN,  NN,  NN,  NN, /* 68 - 6F */
	  NN,  NN,  NN,  NN,  NN,  NN,  NN,  NN, /* 70 - 77 */
	  NN,  NN,  NN,  NN,  NN,  NN,  NN,  NN, /* 78 - 7F */
	  NN,  NN,  NN,  NN,  NN,  NN,  NN,  NN, /* 80 - 87 */
	  NN,  NN,  NN,  NN,  NN,  NN,  NN,  NN, /* 88 - 8F */
	  NN,  NN,  NN,  NN,  NN,  NN,  NN,  NN, /* 90 - 97 */
	  NN,  NN,  NN,  NN,  NN,  NN,  NN,  NN, /* 98 - 9F */
	  NN,  NN,  NN,  NN,  NN,  NN,  NN,  NN, /* A0 - A7 */
	  NN,  NN,  NN,  NN,  NN,  NN,  NN,  NN, /* A8 - AF */
	  NN,  NN,  NN,  NN,  NN,  NN,  NN,  NN, /* B0 - B7 */
	  NN,  NN,  NN,  NN,  NN,  NN,  NN,  NN, /* B8 - BF */
	  NN,  NN,  NN,  NN,  NN,  NN,  NN,  NN, /* C0 - C7 */
	  NN,  NN,  NN,  NN,  NN,  NN,  NN,  NN, /* C8 - CF */
	  NN,  NN,  NN,  NN,  NN,  NN,  NN,  NN, /* D0 - D7 */
	  NN,  NN,  NN,  NN,  NN,  NN,  NN,  NN, /* D8 - DF */
	  29,  42,  56, 219,  157, 54,  184,220, /* E0 - E7 */
	  NN,  NN,  NN,  NN,  NN,  NN,  NN,  NN, /* E8 - EF */
	  NN,  NN,  NN,  NN,  NN,  NN,  NN,  NN, /* F0 - F7 */
	  NN,  NN,  NN,  NN,  NN,  NN,  NN,  NN, /* F8 - FF */
};

//                          KEYBOARD STATUS DEFINITION
#define	KEYBOARD_NOT_IN_USE 		0
#define	KEYBOARD_IN_USE	  			1
#define	KEYBOARD_IN_SHUTDOWN		2
#define KEYBOARD_UNBLOCK			4
#define KEYBOARD_PRESENT			8

#define KEYBOARD_CONTROL_EP			1
#define KEYBOARD_INTIN_EP			2

//                          KEYBOARD HID DEFINITION
#define	USB_SUBCLASS_HID_BOOT		1
#define	USB_PROTOCOL_KEYBOARD		1

//                          REPORTS EQUIVALENCES
#define HID_SET_REPORT		9
#define HID_REPORT_IN		(1 << 8)
#define HID_REPORT_OUT		(2 << 8)
#define HID_REPORT_FEATURE	(3 << 8)

//                          KEYBOARD LEDS EQUIVALENCES
#define	NUMLOCK_MAP		1
#define	CAPSLOCK_MAP  	2
#define	SCROLLLOCK_MAP	4

extern buffer_t *buff_create( unsigned mode, unsigned size, unsigned rsize );
extern int buff_destroy( buffer_t *bptr );
extern int buff_append( buffer_t *bptr, char *dptr, int n );
extern int buff_delete( buffer_t *bptr, char *dptr, int n );
extern int buff_flush( buffer_t *bptr );
extern int buff_putc( buffer_t *bptr, char c );
extern int buff_getc( buffer_t *bptr );
extern int buff_waiting( buffer_t *bptr );
