/*
 * (c) 2010-2019, SWD Embedded Systems Limited, http://www.kpda.ru
 */

// Module Description:  mouse class driver

#include <mouse.h>

mouse_ctrl_t	MouseCtrl;

int mouse_free( mouse_t *mouse )
{
	int		status;

	mouse->status	&= ~( MOUSE_PRESENT | MOUSE_ON );

	if( mouse->ep_int ) {						// Stop the Interrupt In endpoint
		usbd_abort_pipe( mouse->ep_int );
	}

	pthread_sleepon_lock( );
	pthread_sleepon_signal( mouse->rbuf );
	pthread_sleepon_unlock( );

	if( status = rsrcdbmgr_devno_detach( mouse->attr.rdev, 0 ) ) {
		fprintf( stderr, "mouse_free:  rsrcddbmgr_devno_detach %d\n", errno );
	}
	if( status = resmgr_detach( MouseCtrl.dispatch, mouse->resmgr_id, _RESMGR_DETACH_ALL | _RESMGR_DETACH_CLOSE ) ) {
		fprintf( stderr, "mouse_free:  resmgr_detach %d\n", status );
	}
	usbd_free_urb( mouse->urb );
	usbd_free( mouse->ep_int_buf );
	buff_destroy( mouse->rbuf );
	TAILQ_REMOVE( &MouseCtrl.dlist, mouse, dlink );
	if( status = usbd_detach( mouse->device ) ) {
		fprintf( stderr, "mouse_free:  usbd_detach %x\n", status );
	}
	return( status );
}

void mouse_signal_handler( )
{
	siginfo_t		info;
	sigset_t		set;
	mouse_t			*mouse;
	mouse_t			*nmouse;

	sigfillset( &set );
	sigdelset( &set, SIGTERM );
	sigprocmask( SIG_BLOCK, &set, NULL );

	sigemptyset( &set );
	sigaddset( &set, SIGTERM );
	while( SignalWaitinfo( &set, &info ) == -1 )
		;

	for( mouse = TAILQ_FIRST( &MouseCtrl.dlist ); mouse; mouse = nmouse ) {
		nmouse = TAILQ_NEXT( mouse, dlink );
		mouse_free( mouse );
	}
}

int mouse_process( mouse_t *mouse )
{
	uint8_t				mp[3];
	uint8_t				*p;

	// convert report packet to microsoft packet
	p = mouse->ep_int_buf;
	mp[0] = 0x40;
	mp[0] |= (*(p+1) >> 6) & 0x03;
	mp[0] |= (*(p+2) >> 4) & 0x0C;
	mp[0] |= (*(p+0) & 0x01) ? 0x20: 0x00;	// left mouse button
	mp[0] |= (*(p+0) & 0x02) ? 0x10: 0x00;	// right mouse button
	mp[1] = *(p+1) & 0x3F;
	mp[2] = *(p+2) & 0x3F;
//	mp[3] = (*(p+0) & 0x04) ? 0x10: 0x00;	// middle mouse button

	buff_append( mouse->rbuf, (char *)mp, 1 );

	pthread_sleepon_lock( );				// wakeup anyone blocked
	pthread_sleepon_signal( mouse->rbuf );
	pthread_sleepon_unlock( );

	return( EOK );
}

void mouse_int_cbf( struct usbd_urb *urb, struct usbd_pipe *pipe, void *hdl )
{
	mouse_t		*mouse;
	_uint32		ustatus;
	_uint32		alen;

	mouse = (mouse_t *)hdl;

	usbd_urb_status( urb, &ustatus, &alen );

	if( ( ustatus & USBD_URB_STATUS_MASK ) != USBD_STATUS_CMP ) {
		// check for a stall condition and clear it
		if( ( ustatus & USBD_USB_STATUS_MASK ) == USBD_STATUS_STALL ) {
			usbd_reset_pipe( pipe );
		}
		if( MouseCtrl.verbose ) {
			fprintf( stderr, "mouse_int_cbf:  ustatus %x\n", ustatus );
		}
	}
	else {
		mouse_process( mouse );
	}

	if( mouse->status & MOUSE_ON ) {		// re-hook interrupt endpoint
		usbd_setup_interrupt( mouse->urb, URB_DIR_IN|URB_SHORT_XFER_OK, mouse->ep_int_buf, mouse->ep_int_size );
		if( usbd_io( mouse->urb, mouse->ep_int, mouse_int_cbf, mouse, USBD_TIME_INFINITY ) ) {
			pthread_sleepon_lock( );
			mouse->status &= ~( MOUSE_ON | MOUSE_PRESENT );
			pthread_sleepon_signal( mouse->rbuf );
			pthread_sleepon_unlock( );
		}
	}
}

int mouse_io_unblock( resmgr_context_t *ctp, io_pulse_t *msg, RESMGR_OCB_T *ocb )
{
	mouse_t		*mouse;

	mouse			= (mouse_t *)ocb->attr;

	pthread_sleepon_lock( );
	mouse->status	|= MOUSE_UNBLOCK;
	pthread_sleepon_signal( mouse->rbuf );
	pthread_sleepon_unlock( );

	return( _RESMGR_NOREPLY );
}

int mouse_io_open( resmgr_context_t *ctp, io_open_t *msg, RESMGR_HANDLE_T *handle, void *extra )
{
	mouse_t		*mouse;
	int			status;

	mouse	= (mouse_t *)handle;

	if( msg->connect.ioflag & _IO_FLAG_RD ) {
		msg->connect.sflag = SH_DENYRD;
	}

	if( msg->connect.ioflag & _IO_FLAG_WR ) {
		msg->connect.sflag = SH_DENYWR;
	}

	if( status = iofunc_open_default( ctp, msg, handle, extra ) ) {
		return( status );
	}

	if( !( mouse->status & MOUSE_ON ) ) {	// hook callback
		usbd_setup_interrupt( mouse->urb, URB_DIR_IN|URB_SHORT_XFER_OK, mouse->ep_int_buf, mouse->ep_int_size );
		if( status = usbd_io( mouse->urb, mouse->ep_int, mouse_int_cbf, mouse, USBD_TIME_INFINITY ) ) {
			return( status );
		}
		mouse->status |= MOUSE_ON;
	}
	return( EOK );
}

int mouse_io_close( resmgr_context_t *ctp, void *reserved, RESMGR_OCB_T *ocb )
{
	mouse_t		*mouse;

	mouse	= (mouse_t *)ocb->attr;

	iofunc_attr_lock( ocb->attr );
	if( iofunc_ocb_detach( ctp, ocb ) & IOFUNC_OCB_LAST_READER ) {
		mouse->status &= ~MOUSE_ON;
	}
	iofunc_ocb_free( ocb );
	iofunc_attr_unlock( &mouse->attr );
	return( EOK );
}

int mouse_io_read( resmgr_context_t *ctp, io_read_t *msg, RESMGR_OCB_T *ocb )
{
	mouse_t		*mouse;
	buffer_t	*bptr;
	char		mbuf[3];
	int			pos;
	int			nbytes;

	mouse	= (mouse_t *)ocb->attr;
	bptr	= mouse->rbuf;
	pos		= 0;
	nbytes	= msg->i.nbytes;
	mouse->status &= ~MOUSE_UNBLOCK;

	if(	!( mouse->status & MOUSE_PRESENT ) ) {
		return( ENODEV );
	}

	if( ( msg->i.xtype & _IO_XTYPE_MASK ) != _IO_XTYPE_NONE ) {
		return( EINVAL );
	}

	if( !( ocb->ioflag & _IO_FLAG_RD ) ) {
		return( EBADF );
	}

	ocb->flags |= IOFUNC_ATTR_ATIME;

	if( nbytes == 0 ) {
		_IO_SET_READ_NBYTES( ctp, 0 );
		return( EOK );
	}

	while( 1 ) {
// Fix:  optimization buff_delete more than one packet at a time
		while( nbytes >= 3 && buff_delete( bptr, mbuf, 1 ) > 0 ) {
			resmgr_msgwrite( ctp, mbuf, 3, pos );
			nbytes -=3;
			pos += 3;
		}

		if( !nbytes || pos >= mouse->tios.c_cc[VMIN] || nbytes < 3 ) {
			break;
		}
		pthread_sleepon_lock( );
		while( buff_waiting( bptr ) == 0 ) {
			if( mouse->status & MOUSE_UNBLOCK ) {
				pthread_sleepon_unlock( );
				return( EINTR );
			}
			if( !( mouse->status & MOUSE_PRESENT ) ) {
				pthread_sleepon_unlock( );
				return( ENODEV );
			}
			pthread_sleepon_wait( bptr );
		}
		pthread_sleepon_unlock( );
	}

	_IO_SET_READ_NBYTES( ctp, pos );
	return( EOK );
}

int mouse_io_devctl( resmgr_context_t *ctp, io_devctl_t *msg, RESMGR_OCB_T *ocb )
{
	mouse_t		*mouse;
	uint32_t	*data;
	int			nbytes;
	int			status;

	nbytes	= 0;
	data	= (uint32_t *)_DEVCTL_DATA( msg->i );
	mouse	= (mouse_t *)ocb->attr;

	if( !( mouse->status & MOUSE_PRESENT ) ) {
		return( ENODEV );
	}

		// Let common code handle DCMD_ALL_* cases
	if( ( status = iofunc_devctl_default( ctp, msg, ocb ) ) != _RESMGR_DEFAULT ) {
		return( status );
	}

	status = EOK;

	switch( msg->i.dcmd ) {
		case DCMD_CHR_TCGETATTR:
			nbytes = sizeof( struct termios );
			memcpy( data, &mouse->tios, nbytes );
			break;

		case DCMD_CHR_TCSETATTR:
			nbytes = sizeof( struct termios );
			memcpy( &mouse->tios, data, nbytes );
			break;

		case DCMD_CHR_TCFLUSH:
			buff_flush( mouse->rbuf );
			break;

		case DCMD_CHR_TTYINFO: {
			struct _ttyinfo			*ttyinfo;

			ttyinfo = (struct _ttyinfo *)data;
			strcpy( ttyinfo->ttyname, "ser" );
			ttyinfo->opencount = mouse->attr.count;
			nbytes = sizeof( *ttyinfo );
			break;
		}

		case DCMD_CHR_LINESTATUS:
		case DCMD_CHR_SERCTL:
		default:
			status = ENOSYS;
			break;
	}

	msg->o.ret_val = status;
	return( _RESMGR_PTR( ctp, &msg->o, sizeof( msg->o ) + nbytes ) );
}

int mouse_get_devnum( )
{
	int				cnt;
	int				match;
	mouse_t			*mouse;

	for( cnt = 0; cnt < 16; cnt++ ) {
		for( match = 0, mouse = TAILQ_FIRST( &MouseCtrl.dlist ); mouse; mouse = TAILQ_NEXT( mouse, dlink ) ) {
			if( minor( mouse->devnum ) == cnt ) {
				match = 1;
				break;
			}
		}
		if( !match ) {
			break;
		}
	}
	return( cnt );
}

int mouse_io_lock_ocb( resmgr_context_t *ctp, void *reserved, RESMGR_OCB_T *ocb )
{
	return( EOK );
}

int mouse_io_unlock_ocb( resmgr_context_t *ctp, void *reserved, RESMGR_OCB_T *ocb )
{
	return( EOK );
}

int mouse_resmgr_attach( mouse_t *mouse )
{
	char		name[255];

	// setup termios structure
	mouse->tios.c_cc[VMIN] = 3;
	iofunc_attr_init( &mouse->attr, 0600 | S_IFCHR, 0, 0 );
	iofunc_func_init( _RESMGR_CONNECT_NFUNCS, &mouse->resmgr_connect_funcs, _RESMGR_IO_NFUNCS, &mouse->resmgr_io_funcs );
	if( ( mouse->attr.rdev = rsrcdbmgr_devno_attach( "usb", -1, 0 ) ) != -1 ) {
		mouse->devnum = mouse_get_devnum( );
		sprintf( name, "/dev/%s%d", MouseCtrl.prefix, mouse->devnum );
		mouse->resmgr_connect_funcs.open	= mouse_io_open;
		mouse->resmgr_io_funcs.close_ocb	= mouse_io_close;
		mouse->resmgr_io_funcs.unlock_ocb	= mouse_io_unlock_ocb;
		mouse->resmgr_io_funcs.lock_ocb		= mouse_io_lock_ocb;
		mouse->resmgr_io_funcs.read			= mouse_io_read;
		mouse->resmgr_io_funcs.devctl		= mouse_io_devctl;
		mouse->resmgr_io_funcs.unblock		= mouse_io_unblock;
		if( ( mouse->resmgr_id = resmgr_attach( MouseCtrl.dispatch, NULL, name, _FTYPE_ANY, 0,
				&mouse->resmgr_connect_funcs, &mouse->resmgr_io_funcs, &mouse->attr ) ) != -1 ) {
			resmgr_devino( mouse->resmgr_id, &mouse->io_mount.dev, &mouse->attr.inode );
			if( ( mouse->rbuf = buff_create( BUFF_OVERWRITE, 3 * 64, 3 ) ) ) {
				return( EOK );
			}
			else {
				fprintf( stderr, "Mouse:  cannot alloc ring buffer\n" );
			}
		}
		else {
			fprintf( stderr, "Mouse:  resmgr_attach\n" );
		}
		rsrcdbmgr_devno_detach( mouse->attr.rdev, 0 );
	}
	else {
		fprintf( stderr, "Unable to acquire devno\n" );
	}

	return( EIO );
}

int mouse_parse_descriptors( mouse_t *mouse )
{
	_uint32						eix;
	_uint32						scan;
	_uint32						found;
	usbd_descriptors_t			*desc;
	struct usbd_desc_node		*ifc, *ept;

	scan	= MOUSE_CONTROL_EP | MOUSE_INTIN_EP;
	found	= 0;

	if( usbd_interface_descriptor( mouse->device, mouse->instance.config, mouse->instance.iface, mouse->instance.alternate, &ifc ) ) {
		for( eix = 0; ( desc = usbd_parse_descriptors( mouse->device, ifc, USB_DESC_ENDPOINT, eix, &ept ) ) != NULL; ++eix ) {
			switch( desc->endpoint.bmAttributes ) {
				case USB_ATTRIB_CONTROL:
					if( usbd_open_pipe( mouse->device, desc, &mouse->ep_cntl ) == EOK ) {
						found |= MOUSE_CONTROL_EP;
					}
					break;

				case USB_ATTRIB_ISOCHRONOUS:
					break;

				case USB_ATTRIB_BULK:
					break;

				case USB_ATTRIB_INTERRUPT:
					switch( desc->endpoint.bEndpointAddress & USB_ENDPOINT_IN ) {
						case USB_ENDPOINT_OUT:
							break;

						case USB_ENDPOINT_IN:
							if( usbd_open_pipe( mouse->device, desc,
									&mouse->ep_int ) == EOK ) {
								mouse->ep_int_size = desc->endpoint.wMaxPacketSize;
								found |= MOUSE_INTIN_EP;
							}
							break;
					}
					break;
			}
		}
	}
	return( ( found == scan ) ? EOK : ENODEV );
}

void mouse_insertion( struct usbd_connection *connection, usbd_device_instance_t *instance )
{
	mouse_t						*mouse;
	struct usbd_device			*device;
	int							status;

	if( ( status = usbd_attach( connection, instance, sizeof( mouse_t ), &device ) ) == EOK ) {
		mouse				= usbd_device_extra( device );
		mouse->device		= device;
		mouse->instance		= *instance;
		if( mouse_parse_descriptors( mouse ) == EOK ) {
			if( mouse->ep_int_buf = usbd_alloc( mouse->ep_int_size ) ) {
				if( mouse->urb = usbd_alloc_urb( NULL ) ) {
					if( mouse_resmgr_attach( mouse ) == EOK ) {
						mouse->status = MOUSE_PRESENT;
						TAILQ_INSERT_TAIL( &MouseCtrl.dlist, mouse, dlink );
						return;
					}
					usbd_free_urb( mouse->urb );
				}
				usbd_free( mouse->ep_int_buf );
			}
		}
		usbd_detach( device );
	}
}

void mouse_removal( struct usbd_connection *connection, usbd_device_instance_t *instance )
{
	struct usbd_device	*device;
	mouse_t				*mouse;

	if( ( device = usbd_device_lookup( connection, instance ) ) != NULL ) {
		mouse = usbd_device_extra( device );
		mouse_free( mouse );
	}
}

int mouse_resmgr_init( )
{
	resmgr_attr_t		attr;

	if( procmgr_daemon( EXIT_SUCCESS, PROCMGR_DAEMON_NOCLOSE | PROCMGR_DAEMON_NODEVNULL ) == -1 ) {
		fprintf( stderr, "Mouse:  procmgr_daemon\n" );
		return( -1 );
	}

	if( ( MouseCtrl.dispatch = dispatch_create( ) ) == NULL ) {
		fprintf( stderr, "Unable to allocate dispatch context\n" );
		return( -1 );
	}

	memset( &MouseCtrl.pool_attr, 0, sizeof( MouseCtrl.pool_attr ) );
	if( ( MouseCtrl.pool_attr.attr = malloc( sizeof( *MouseCtrl.pool_attr.attr ) ) ) == NULL ) {
		fprintf( stderr, "Unable to alloc pool attr\n" );
		return( -1 );
	}

	pthread_attr_init( MouseCtrl.pool_attr.attr );
	pthread_attr_setstacksize( MouseCtrl.pool_attr.attr, 8192 );

	MouseCtrl.pool_attr.handle			= MouseCtrl.dispatch;
	MouseCtrl.pool_attr.context_alloc	= resmgr_context_alloc;
	MouseCtrl.pool_attr.block_func		= resmgr_block;
	MouseCtrl.pool_attr.handler_func	= (void *)dispatch_handler;
	MouseCtrl.pool_attr.context_free	= resmgr_context_free;
	MouseCtrl.pool_attr.lo_water		= 1;
	MouseCtrl.pool_attr.hi_water		= 2;
	MouseCtrl.pool_attr.increment		= 1;
	MouseCtrl.pool_attr.maximum			= 2;

	if( ( MouseCtrl.thread_pool = thread_pool_create( &MouseCtrl.pool_attr, 0 ) ) == NULL ) {
		fprintf( stderr, "Unable to initialize thread pool\n" );
		return( -1 );
	}

	memset( &attr, 0, sizeof( attr ) );
	attr.nparts_max		= 10;
	attr.msg_max_size	= 0;

	if( resmgr_attach( MouseCtrl.dispatch, &attr, NULL, 0, 0, NULL, NULL, NULL ) == -1 ) {
		fprintf( stderr, "Unable to attach name\n" );
		return( -1 );
	}

	if( thread_pool_start( MouseCtrl.thread_pool ) ) {
		fprintf( stderr, "Unable to initialize thread_pool\n" );
		return( -1 );
	}

	return( 0 );
}

int mouse_start_driver( int argc, char *argv[] )
{
	struct usbd_connection	*connection;
	usbd_funcs_t			funcs = { _USBDI_NFUNCS, mouse_insertion, mouse_removal, NULL };
	usbd_device_ident_t		interest = { USBD_CONNECT_WILDCARD, USBD_CONNECT_WILDCARD, USB_CLASS_HID, USB_SUBCLASS_HID_BOOT, USB_PROTOCOL_MOUSE };
	usbd_connect_parm_t		parm = { NULL, USB_VERSION, USBD_VERSION, 0, argc, argv, 0, &interest, &funcs, USBD_CONNECT_WAIT };
	int						status;
	int						opt;
	sigset_t				signals;

	sigfillset( &signals );					// No signals to resmgr threads
	pthread_sigmask( SIG_BLOCK, &signals, NULL );

	TAILQ_INIT( &MouseCtrl.dlist );
	MouseCtrl.prefix = "usbmouse";

	while( ( opt = getopt( argc, argv, "vn:s:w:" ) ) != -1 ) {
		switch( opt ) {
			case 'n':
				MouseCtrl.prefix = optarg;
				break;

			case 'v':
				MouseCtrl.verbose++;
				break;

			case 's':
				parm.path = optarg;
				break;

			case 'w':
				parm.connect_wait = strtol( optarg, 0, 0 );
				break;

			default:
				break;
		}
	}

	if( ( MouseCtrl.chid = ChannelCreate( _NTO_CHF_DISCONNECT | _NTO_CHF_UNBLOCK ) ) == -1 ||
	   ( MouseCtrl.coid = ConnectAttach( 0, 0, MouseCtrl.chid, _NTO_SIDE_CHANNEL, 0 ) ) == -1 ) {
		fprintf( stderr, "Unable to attach channel and connection\n" );
		return( errno );
	}

	if( mouse_resmgr_init( ) ) {
		fprintf( stderr, "Mouse:  Resmgr Init Failure\n" );
		return( errno );
	}

	if( ( status = usbd_connect( &parm, &connection ) ) != EOK ) {
		fprintf( stderr, "Mouse Class/init - %s\n", strerror( status ) );
		return( status );
	}
	return( status );
}

int main( int argc, char *argv[] )
{
	if( mouse_start_driver( argc, argv ) ) {
		exit( EXIT_FAILURE );
	}
	mouse_signal_handler( );
	exit( EXIT_SUCCESS );
}
