/*
 * (c) 2010-2019, SWD Embedded Systems Limited, http://www.kpda.ru
 */

// Module Description:  printer class driver

#include <printer.h>

prn_ctrl_t				PrnCtrl;

int prn_free( prn_t *prn )
{
	int		status;

	if( prn->ep_out ) {						// Stop the Bulk Out endpoint
		usbd_abort_pipe( prn->ep_out );
	}

	if( prn->ep_in ) {						// Stop the Bulk In endpoint
		usbd_abort_pipe( prn->ep_in );
	}

	pthread_sleepon_lock( );
	prn->status	&= ~PRN_PRESENT;
	pthread_sleepon_signal( &prn->urb );
	pthread_sleepon_unlock( );

	if( status = rsrcdbmgr_devno_detach( prn->attr.rdev, 0 ) ) {
		fprintf( stderr, "prn_free:  rsrcddbmgr_devno_detach %d\n", errno );
	}
	if( status = resmgr_detach( PrnCtrl.dispatch, prn->resmgr_id, _RESMGR_DETACH_ALL | _RESMGR_DETACH_CLOSE ) ) {
		fprintf( stderr, "prn_free:  resmgr_detach %d\n", status );
	}
	usbd_free( prn->buffer );
	usbd_free_urb( prn->urb );
	TAILQ_REMOVE( &PrnCtrl.dlist, prn, dlink );
	if( status = usbd_detach( prn->device ) ) {
		fprintf( stderr, "prn_free:  pending I/O\n" );
	}
	return( status );
}

void prn_signal_handler( )
{
	siginfo_t		info;
	sigset_t		set;
	prn_t			*prn;
	prn_t			*nprn;

	sigfillset( &set );
	sigdelset( &set, SIGTERM );
	sigprocmask( SIG_BLOCK, &set, NULL );

	sigemptyset( &set );
	sigaddset( &set, SIGTERM );
	while( SignalWaitinfo( &set, &info ) == -1 )
		;

	for( prn = TAILQ_FIRST( &PrnCtrl.dlist ); prn; prn = nprn ) {
		nprn = TAILQ_NEXT( prn, dlink );
		prn_free( prn );
	}
}

void prn_bulk_cbf( struct usbd_urb *urb, struct usbd_pipe *pipe, void *hdl )
{
	pipe = pipe, hdl = hdl;

	pthread_sleepon_lock( );
	pthread_sleepon_signal( urb );
	pthread_sleepon_unlock( );

	return;
}

int prn_bulk_wait( prn_t *prn, struct usbd_urb *urb, struct usbd_pipe *pipe )
{
	_uint32		ustatus;
	_uint32		alen;

	pthread_sleepon_lock( );
	while( usbd_urb_status( urb, &ustatus, &alen ) == EBUSY &&
				!( prn->status & PRN_UNBLOCK ) ) {
		pthread_sleepon_wait( urb );
	}
	pthread_sleepon_unlock( );

	// check for a stall condition and clear it
	if( ( ustatus & USBD_USB_STATUS_MASK ) == USBD_STATUS_STALL ) {
		usbd_reset_pipe( pipe );
	}

	if( prn->status & PRN_UNBLOCK ) {
		return( EINTR );
	}

	return( ( ustatus & USBD_URB_STATUS_MASK ) == USBD_STATUS_CMP ? EOK : EIO );
}

int prn_io( prn_t *prn, struct usbd_pipe *pipe, _uint32 flags, _uint8 *buffer, _uint32 length )
{
	_uint32		status;

	usbd_setup_bulk( prn->urb, flags, buffer, length );
	if( ( status = usbd_io( prn->urb, pipe, prn_bulk_cbf, prn, USBD_TIME_INFINITY ) ) == EOK ) {
		return( prn_bulk_wait( prn, prn->urb, pipe ) );
	}
	return( status );
}

int prn_device_id( prn_t *prn, char *buffer, uint32_t length )
{
	_uint32			status;
	struct usbd_urb	*urb;

	if( urb = usbd_alloc_urb( NULL ) ) {
		usbd_setup_vendor( urb, URB_DIR_IN, 0, USB_TYPE_CLASS | USB_RECIPIENT_INTERFACE, 0, 0, buffer, length );
		if( ( status = usbd_io( urb, prn->ep_cntl, NULL, prn, USBD_TIME_INFINITY ) ) == EOK ) {
		}
		usbd_free_urb( urb );
	}
	else {
		status = ENOMEM;
	}

	return( status );
}

int prn_reset( prn_t *prn )
{
	_uint32			status;
	struct usbd_urb	*urb;

// Note:  version 1.0 of the spec incorrectly stated that
// the recipient should be other, version 1.1 states that
// it should be interface

	status = ENOMEM;
	if( urb = usbd_alloc_urb( NULL ) ) {
		usbd_setup_vendor( urb, URB_DIR_OUT, PRN_SOFT_RESET, USB_TYPE_CLASS |
			( ( prn->bcdUSB & 0x10 ) ? USB_RECIPIENT_INTERFACE : USB_RECIPIENT_OTHER ),
				0, 0, NULL, 0 );
		if( ( status = usbd_io( urb, prn->ep_cntl, NULL, prn, USBD_TIME_INFINITY ) ) == EOK ) {
		}
		usbd_free_urb( urb );
	}

	return( status );
}

int prn_status( prn_t *prn )
{
	_uint32			status;
	struct usbd_urb	*urb;
	_uint8			*buffer;

	status = ENOMEM;
	if( urb = usbd_alloc_urb( NULL ) ) {
		if( buffer = usbd_alloc( 1 ) ) {
			usbd_setup_vendor( urb, URB_DIR_IN, 1, USB_TYPE_CLASS | USB_RECIPIENT_INTERFACE, 0, 0, buffer, 1 );
			if( ( status = usbd_io( urb, prn->ep_cntl, NULL, prn, USBD_TIME_INFINITY ) ) == EOK ) {
				prn->query_status = ( prn->query_status & ~0xff ) | *buffer;
			}
			usbd_free( buffer );
		}
		usbd_free_urb( urb );
	}

	return( status );
}

int prn_io_unblock( resmgr_context_t *ctp, io_pulse_t *msg, RESMGR_OCB_T *ocb )
{
	prn_t		*prn;

	prn	= (prn_t *)ocb->attr;

	pthread_sleepon_lock( );
	prn->status |= PRN_UNBLOCK;
	pthread_sleepon_signal( prn->urb );
	pthread_sleepon_unlock( );
	
	return( _RESMGR_NOREPLY );
}

int prn_io_open( resmgr_context_t *ctp, io_open_t *msg, RESMGR_HANDLE_T *handle, void *extra )
{
	prn_t		*prn;
	int			status;

	prn	= (prn_t *)handle;

	if( !( prn->status & PRN_PRESENT ) ) {
		return( ENODEV );
	}

	if( status = iofunc_open_default( ctp, msg, handle, extra ) ) {
		return( status );
	}

	return( EOK );
}

int prn_io_close( resmgr_context_t *ctp, void *reserved, RESMGR_OCB_T *ocb )
{
	prn_t		*prn;

	prn	= (prn_t *)ocb->attr;

	iofunc_attr_lock( ocb->attr );
	if( iofunc_ocb_detach( ctp, ocb ) & IOFUNC_OCB_LAST_WRITER ) {
		// terminate the printer transfer with a null Bulk Out packet
		if( prn->status & PRN_PRESENT ) {
			if( prn_io( prn, prn->ep_out, URB_DIR_OUT, prn->buffer, 0 ) ) {
			}
		}
	}
	iofunc_ocb_free( ocb );
	iofunc_attr_unlock( &prn->attr );
	return( EOK );
}

int prn_io_read( resmgr_context_t *ctp, io_read_t *msg, RESMGR_OCB_T *ocb )
{
	prn_t		*prn;
	_uint32		pos;
	_uint8		*ibuf;
	_uint32		nbytes;
	_uint32		status;
	_uint32		alen;

	prn	= (prn_t *)ocb->attr;
	pos		= 0;
	ibuf	= prn->buffer;
	nbytes	= msg->i.nbytes;

	if( !( prn->status & PRN_PRESENT ) ) {
		return( ENODEV );
	}

	if( prn->protocol != USB_PROTOCOL_PRN_BIDIR ) {
		return( ENXIO );
	}

	if( ( msg->i.xtype & _IO_XTYPE_MASK ) != _IO_XTYPE_NONE ) {
		return( EINVAL );
	}

	if( !( ocb->ioflag & _IO_FLAG_RD ) ) {
		return( EBADF );
	}

	ocb->flags |= IOFUNC_ATTR_ATIME;

	if( nbytes == 0 ) {
		_IO_SET_READ_NBYTES( ctp, 0 );
		return( EOK );
	}

	if( atomic_set_value( &prn->status, PRN_READING ) & PRN_READING ) {
		return( EBUSY );
	}

	prn->status &= ~PRN_UNBLOCK;
	while( 1 ) {
		if( status = prn_io( prn, prn->ep_in, URB_DIR_IN, ibuf, min( nbytes, MAX_PRN_BUFFER ) ) ) {
			break;
		}

		usbd_urb_status( prn->urb, NULL, &alen );
		resmgr_msgwrite( ctp, ibuf, alen, pos );
		nbytes	-= alen;
		pos		+= alen;

		if( prn->status & PRN_UNBLOCK ) {
			status = EINTR;
			break;
		}

		if( !nbytes || !alen ) {
			atomic_clr_value( &prn->status, PRN_READING );
			_IO_SET_READ_NBYTES( ctp, pos );
			return( EOK );
		}
	}

	atomic_clr_value( &prn->status, PRN_READING );
	return( status );
}

int prn_io_write( resmgr_context_t *ctp, io_write_t *msg, RESMGR_OCB_T *ocb )
{
	prn_t		*prn;
	uint8_t		*ibuf;
	uint8_t		*obuf;
	int			off, iidx, oidx;
	int 		rlen;
	int 		ilen;
	int			opost;
	int			olen;
	int			status;

	prn	= (prn_t *)ocb->attr;
	rlen	= msg->i.nbytes;

	if( !( prn->status & PRN_PRESENT ) ) {
		return( ENODEV );
	}

	if( ( msg->i.xtype & _IO_XTYPE_MASK ) != _IO_XTYPE_NONE ) {
		return( EINVAL );
	}

	if( !( ocb->ioflag & _IO_FLAG_WR ) ) {
		return( EBADF );
	}

	ocb->flags |= IOFUNC_ATTR_ATIME;

	if( rlen == 0 ) {
		_IO_SET_WRITE_NBYTES( ctp, 0 );
		return( EOK );
	}

	if( atomic_set_value( &prn->status, PRN_WRITING ) & PRN_WRITING ) {
		return( EBUSY );
	}

	prn->status &= ~PRN_UNBLOCK;
	opost	= prn->tios.c_oflag & OPOST;
	off		= sizeof( io_write_t );
	ibuf	= prn->buffer;
	obuf	= ibuf + ( opost ? 512 : 0 );
	while( rlen ) {
		ilen = ( rlen <= MAX_PRN_BUFFER ) ? rlen : MAX_PRN_BUFFER;
		if( ( olen = resmgr_msgread( ctp, ibuf, ilen, off ) ) != ilen ) {
			break;
		}

		if( opost ) {
			for( iidx = 0, oidx = 0; iidx < ilen; iidx++, oidx++ ) {
				if( ibuf[iidx] == 0xa ) {
					obuf[oidx++] = 0xd;
					olen++;
				}
				obuf[oidx] = ibuf[iidx];
			}
		}

		if( status = prn_io( prn, prn->ep_out, URB_DIR_OUT, obuf, olen ) ) {
			atomic_clr_value( &prn->status, PRN_WRITING );
			return( status );
		}
		rlen -= ilen;
		off += MAX_PRN_BUFFER;
	}

	atomic_clr_value( &prn->status, PRN_WRITING );
	_IO_SET_WRITE_NBYTES( ctp, msg->i.nbytes - rlen );
	return( EOK );
}

int prn_io_devctl( resmgr_context_t *ctp, io_devctl_t *msg, RESMGR_OCB_T *ocb )
{
	prn_t		*prn;
	uint32_t	*data;
	int			nbytes;
	int			status;

	nbytes	= 0;
	data	= (uint32_t *)_DEVCTL_DATA( msg->i );
	prn	= (prn_t *)ocb->attr;

	if( !( prn->status & PRN_PRESENT ) ) {
		return( ENODEV );
	}

		// Let common code handle DCMD_ALL_* cases
	if( ( status = iofunc_devctl_default( ctp, msg, ocb ) ) != _RESMGR_DEFAULT ) {
		return( status );
	}

	status = EOK;

	switch( msg->i.dcmd ) {
		case DCMD_CHR_TCGETATTR:
			nbytes = sizeof( struct termios );
			memcpy( data, &prn->tios, nbytes );
			break;

		case DCMD_CHR_TCSETATTR:
			nbytes = sizeof( struct termios );
			memcpy( &prn->tios, data, nbytes );
			break;

		case DCMD_CHR_TCFLUSH:
			break;

		case DCMD_CHR_TTYINFO: {
			struct _ttyinfo			*ttyinfo;

			ttyinfo = (struct _ttyinfo *)data;
			sprintf( ttyinfo->ttyname, "/dev/%s%d", PrnCtrl.prefix, prn->devnum );
			ttyinfo->opencount = prn->attr.count;
			nbytes = sizeof( *ttyinfo );
			break;
		}

		case DCMD_CHR_LINESTATUS:
			prn_status( prn );
			*data = prn->query_status;
			nbytes = sizeof( *data );
			break;

		case DCMD_CHR_PARCTL:
			if( *data & 0x4 ) {
				status = prn_reset( prn );
			}
			break;

		case DCMD_CHR_SERCTL:
			break;

		case DCMD_CHR_PNPTEXT: {
			char *id;

			if( id = usbd_alloc( MAX_PRN_DEVICE_NAME ) ) {
				prn_device_id( prn, id, MAX_PRN_DEVICE_NAME );
				nbytes = min( ENDIAN_BE16( *((uint16_t *)id ) ), msg->i.nbytes );
				strncpy( (void *)data, id + 2, nbytes );
				usbd_free( id );
			}
			break;
		}

		default:
			status = ENOSYS;
			break;
	}

	msg->o.ret_val = status;
	return( _RESMGR_PTR( ctp, &msg->o, sizeof( msg->o ) + nbytes ) );
}

int prn_get_devnum( )
{
	int				cnt;
	int				match;
	prn_t			*prn;

	for( cnt = 0; cnt < 16; cnt++ ) {
		for( match = 0, prn = TAILQ_FIRST( &PrnCtrl.dlist ); prn; prn = TAILQ_NEXT( prn, dlink ) ) {
			if( minor( prn->devnum ) == cnt ) {
				match = 1;
				break;
			}
		}
		if( !match ) {
			break;
		}
	}
	return( cnt );
}

int prn_io_lock_ocb( resmgr_context_t *ctp, void *reserved, RESMGR_OCB_T *ocb )
{
	return( EOK );
}

int prn_io_unlock_ocb( resmgr_context_t *ctp, void *reserved, RESMGR_OCB_T *ocb )
{
	return( EOK );
}

int prn_resmgr_attach( prn_t *prn )
{
	char		name[255];

	// setup termios structure
	prn->tios.c_oflag = 0; // OPOST;
	prn->query_status = 0x418;
	iofunc_attr_init( &prn->attr, 0600 | S_IFCHR, 0, 0 );
	iofunc_func_init( _RESMGR_CONNECT_NFUNCS, &prn->resmgr_connect_funcs, _RESMGR_IO_NFUNCS, &prn->resmgr_io_funcs );
	if( ( prn->attr.rdev = rsrcdbmgr_devno_attach( "usb", -1, 0 ) ) != -1 ) {
		prn->devnum = prn_get_devnum( );
		sprintf( name, "/dev/%s%d", PrnCtrl.prefix, prn->devnum );
		prn->resmgr_connect_funcs.open		= prn_io_open;
		prn->resmgr_io_funcs.close_ocb		= prn_io_close;
		prn->resmgr_io_funcs.unlock_ocb		= prn_io_unlock_ocb;
		prn->resmgr_io_funcs.lock_ocb		= prn_io_lock_ocb;
		prn->resmgr_io_funcs.read			= prn_io_read;
		prn->resmgr_io_funcs.write			= prn_io_write;
		prn->resmgr_io_funcs.devctl			= prn_io_devctl;
		prn->resmgr_io_funcs.unblock		= prn_io_unblock;
		if( ( prn->resmgr_id = resmgr_attach( PrnCtrl.dispatch, NULL, name, _FTYPE_ANY, 0,
				&prn->resmgr_connect_funcs, &prn->resmgr_io_funcs, &prn->attr ) ) != -1 ) {
			resmgr_devino( prn->resmgr_id, &prn->io_mount.dev, &prn->attr.inode );
			return( EOK );
		}
		else {
			fprintf( stderr, "PRN:  resmgr_attach\n" );
		}
		rsrcdbmgr_devno_detach( prn->attr.rdev, 0 );
	}
	else {
		fprintf( stderr, "Unable to acquire devno\n" );
	}

	return( EIO );
}

int prn_parse_descriptors( prn_t *prn )
{
	_uint32						eix;
	_uint32						scan;
	_uint32						found;
	usbd_descriptors_t			*desc;
	usbd_device_descriptor_t	*ddesc;
	usbd_interface_descriptor_t	*idesc;
	struct usbd_desc_node		*dev, *ifc, *ept;
	int							alternate;

	scan	= PRN_CONTROL_EP | PRN_BULKOUT_EP;
	found	= 0;

	if( ddesc = usbd_device_descriptor( prn->device, &dev ) ) {
		prn->bcdUSB = ddesc->bcdUSB;
		// look for interface with bi-directional or uni-directional protocol (No support  for IEEE 1284.4 multi function devices)
		for ( alternate = 0 ; ( idesc = usbd_interface_descriptor( prn->device, prn->instance.config, prn->instance.iface, alternate, &ifc ) ) ; alternate++ ) {
			if( (idesc->bInterfaceProtocol != USB_PROTOCOL_PRN_BIDIR) && (idesc->bInterfaceProtocol != USB_PROTOCOL_PRN_UNIDIR)  ) 
				continue;
				
			if( ( prn->protocol = idesc->bInterfaceProtocol ) == USB_PROTOCOL_PRN_BIDIR ) {
				scan |= PRN_BULKIN_EP;
			}
			usbd_select_interface( prn->device, prn->instance.iface, alternate );
			for( eix = 0; ( desc = usbd_parse_descriptors( prn->device, ifc, USB_DESC_ENDPOINT, eix, &ept ) ) != NULL; ++eix ) {
				switch( desc->endpoint.bmAttributes ) {
					case USB_ATTRIB_CONTROL:
						if( usbd_open_pipe( prn->device, desc, &prn->ep_cntl ) == EOK ) {
							found |= PRN_CONTROL_EP;
						}
						break;

					case USB_ATTRIB_ISOCHRONOUS:
						break;

					case USB_ATTRIB_BULK:
						switch( desc->endpoint.bEndpointAddress & USB_ENDPOINT_IN ) {
							case USB_ENDPOINT_OUT:
								if( usbd_open_pipe( prn->device, desc,
									&prn->ep_out ) == EOK ) {
									found |= PRN_BULKOUT_EP;
								}
								break;

							case USB_ENDPOINT_IN:
								if( ( scan & PRN_BULKIN_EP ) &&
										usbd_open_pipe( prn->device, desc,
										&prn->ep_in ) == EOK ) {
									found |= PRN_BULKIN_EP;
								}
								break;
						}
						break;

					case USB_ATTRIB_INTERRUPT:
						break;
				}
			}
			if( found == scan )
				break;
		}
	}

	return( ( found == scan ) ? EOK : ENODEV );
}

void prn_insertion( struct usbd_connection *connection, usbd_device_instance_t *instance )
{
	prn_t						*prn;
	struct usbd_device			*device;
	int							status;
	if( ( status = usbd_attach( connection, instance, sizeof( prn_t ), &device ) ) == EOK ) {
		prn				= usbd_device_extra( device );
		prn->device		= device;
		prn->instance	= *instance;
		if( prn_parse_descriptors( prn ) == EOK ) {
			if( prn->buffer	= usbd_alloc( PRN_BUFFER_SIZE ) ) {
				if( prn->urb = usbd_alloc_urb( NULL ) ) {
					if( prn_resmgr_attach( prn ) == EOK ) {
						prn->status = PRN_PRESENT;
						TAILQ_INSERT_TAIL( &PrnCtrl.dlist, prn, dlink );
						return;
					}
					usbd_free_urb( prn->urb );
				}
				usbd_free( prn->buffer );
			}
		}
		usbd_detach( device );
	}
}

void prn_removal( struct usbd_connection *connection, usbd_device_instance_t *instance )
{
	struct usbd_device	*device;
	prn_t				*prn;

	if( ( device = usbd_device_lookup( connection, instance ) ) != NULL ) {
		prn = usbd_device_extra( device );
		prn_free( prn );
	}
}

int prn_resmgr_init( )
{
	resmgr_attr_t		attr;

	if( procmgr_daemon( EXIT_SUCCESS, PROCMGR_DAEMON_NOCLOSE | PROCMGR_DAEMON_NODEVNULL ) == -1 ) {
		fprintf( stderr, "Prn:  procmgr_daemon\n" );
		return( -1 );
	}

	if( ( PrnCtrl.dispatch = dispatch_create( ) ) == NULL ) {
		fprintf( stderr, "Unable to allocate dispatch context\n" );
		return( -1 );
	}

	memset( &PrnCtrl.pool_attr, 0, sizeof( PrnCtrl.pool_attr ) );
	if( ( PrnCtrl.pool_attr.attr = malloc( sizeof( *PrnCtrl.pool_attr.attr ) ) ) == NULL ) {
		fprintf( stderr, "Unable to alloc pool attr\n" );
		return( -1 );
	}

	pthread_attr_init( PrnCtrl.pool_attr.attr );
	pthread_attr_setstacksize( PrnCtrl.pool_attr.attr, 8192 );

	PrnCtrl.pool_attr.handle		= PrnCtrl.dispatch;
	PrnCtrl.pool_attr.context_alloc	= resmgr_context_alloc;
	PrnCtrl.pool_attr.block_func	= resmgr_block;
	PrnCtrl.pool_attr.handler_func	= (void *)dispatch_handler;
	PrnCtrl.pool_attr.context_free	= resmgr_context_free;
	PrnCtrl.pool_attr.lo_water		= 1;
	PrnCtrl.pool_attr.hi_water		= 2;
	PrnCtrl.pool_attr.increment		= 1;
	PrnCtrl.pool_attr.maximum		= 2;

	if( ( PrnCtrl.thread_pool = thread_pool_create( &PrnCtrl.pool_attr, 0 ) ) == NULL ) {
		fprintf( stderr, "Unable to initialize thread pool\n" );
		return( -1 );
	}

	memset( &attr, 0, sizeof( attr ) );
	attr.nparts_max		= 10;
	attr.msg_max_size	= 0;

	if( resmgr_attach( PrnCtrl.dispatch, &attr, NULL, 0, 0, NULL, NULL, NULL ) == -1 ) {
		fprintf( stderr, "Unable to attach name\n" );
		return( -1 );
	}

	if( thread_pool_start( PrnCtrl.thread_pool ) ) {
		fprintf( stderr, "Unable to initialize thread_pool\n" );
		return( -1 );
	}

	return( 0 );
}

int prn_start_driver( int argc, char *argv[] )
{
	struct usbd_connection	*connection;
	usbd_funcs_t			funcs = { _USBDI_NFUNCS, prn_insertion, prn_removal, NULL };
	usbd_device_ident_t		interest = { USBD_CONNECT_WILDCARD, USBD_CONNECT_WILDCARD, USB_CLASS_PRN, USB_SUBCLASS_PRN, USBD_CONNECT_WILDCARD };
	usbd_connect_parm_t		parm = { NULL, USB_VERSION, USBD_VERSION, 0, argc, argv, 0, &interest, &funcs, USBD_CONNECT_WAIT };
	int						status;
	int						opt;
	sigset_t				signals;

	sigfillset( &signals );					// No signals to resmgr threads
	pthread_sigmask( SIG_BLOCK, &signals, NULL );

	TAILQ_INIT( &PrnCtrl.dlist );
	PrnCtrl.prefix = "usbpar";

	while( ( opt = getopt( argc, argv, "vn:s:w:" ) ) != -1 ) {
		switch( opt ) {
			case 'n':
				PrnCtrl.prefix = optarg;
				break;

			case 'v':
				PrnCtrl.verbose++;
				break;

			case 's':
				parm.path = optarg;
				break;

			case 'w':
				parm.connect_wait = strtol( optarg, 0, 0 );
				break;

			default:
				break;
		}
	}

	if( ( PrnCtrl.chid = ChannelCreate( _NTO_CHF_DISCONNECT | _NTO_CHF_UNBLOCK ) ) == -1 ||
	   ( PrnCtrl.coid = ConnectAttach( 0, 0, PrnCtrl.chid, _NTO_SIDE_CHANNEL, 0 ) ) == -1 ) {
		fprintf( stderr, "Unable to attach channel and connection\n" );
		return( errno );
	}

	if( prn_resmgr_init( ) ) {
		fprintf( stderr, "Printer:  Resmgr Init Failure\n" );
		return( errno );
	}

	if( ( status = usbd_connect( &parm, &connection ) ) != EOK ) {
		fprintf( stderr, "Prn Class/init - %s\n", strerror( status ) );
		return( status );
	}

	return( status );
}

int main( int argc, char *argv[] )
{
	if( prn_start_driver( argc, argv ) ) {
		exit( EXIT_FAILURE );
	}
	prn_signal_handler( );
	exit( EXIT_SUCCESS );
}
